UNIT MCB;

{$mode objfpc}{$H+}

INTERFACE

USES
  CONTNRS;

CONST
  idlen      = 32;
  bufsize    = 256;
  maxexp     = 38;
  wordsize   = 4;
  maxint     = 2147483647;
  maxhuge    = 9223372036854775807;
  minint     = - maxint - 1;

  cHead   = 0;  cConst  = 1;  cVars   = 2;  cPar     = 3;
  cFld    = 4;  cTyp    = 5;  cProc   = 6;  cFunc    = 7;
  cMod    = 8;  cNode   = 9;  cLit    = 10;
  fByte   = 1;  fBool   = 2;  fChar   = 3;  fInt     = 4;
  fHuge   = 5;  fReal   = 6;  fSet    = 7;  fPointer = 8;
  fNilTyp = 9;  fNoTyp  = 10; fProc   = 11; fString  = 12;
  fArray  = 13; fRecord = 14; fCArray = 15;

TYPE
  Ident   = ARRAY [0..idlen-1]   OF CHAR;
  Strings = ARRAY [0..bufsize-1] OF CHAR;

  Object_ = CLASS;

  Type_   = CLASS
    base  : Type_;
    typobj: Object_;
    dsc   : Object_;
    len   : LONGINT;
    size  : LONGINT;
    form  : LONGINT;
  END;

  Item    = CLASS
    name   : Ident;
    typ    : Type_;
    a,b    : Item;
    op     : INTEGER;
    clas   : INTEGER;
    val    : INTEGER;
    reg    : INTEGER;
  END;

  Object_ = CLASS (Item)
    exno   : INTEGER;
    dsc    : Object_;
    c      : Object_;
  END;

VAR
  GC       : TObjectList;
  pos      : INTEGER;
  strx     : INTEGER;
  version  : INTEGER;
  errpos   : INTEGER;
  errcount : INTEGER;
  prevcol  : INTEGER;
  prevline : INTEGER;

  Mark     : PROCEDURE (msg: ARRAY OF CHAR);

  topScope, hidden, dummy  : Object_;
  byteType, boolType, charType, intType, realType : Type_;
  setType,  nilType,  noType,   strType, hugeType : Type_;

  PROCEDURE NEW(VAR x: Type_);
  PROCEDURE NEW(VAR x: Item); OVERLOAD;
  PROCEDURE NEW(VAR x: Object_); OVERLOAD;

  PROCEDURE OpenScope;
  PROCEDURE CloseScope;
  PROCEDURE NewObj(VAR x: Object_; id: Ident; clas: INTEGER);
  FUNCTION  ThisObj(id: Ident): Object_;

  FUNCTION  NewItem(clas: INTEGER; a, b: Item; op: INTEGER): Item;
  PROCEDURE MakeConstItem(VAR x: Item; typ: Type_; val: LONGINT);

IMPLEMENTATION

PROCEDURE NEW(VAR x: Type_);   BEGIN x := Type_.Create; GC.Add(x);   END;
PROCEDURE NEW(VAR x: Item);    BEGIN x := Item.Create; GC.Add(x);    END;
PROCEDURE NEW(VAR x: Object_); BEGIN x := Object_.Create; GC.Add(x); END;

PROCEDURE MarkProc(msg: ARRAY OF CHAR);
VAR
  p: LONGINT;
BEGIN
  p := pos;
  IF errcount = 0 THEN BEGIN END;
  IF (p > pos) AND (errcount < 25) THEN BEGIN END;
  INC(errcount);
  pos := p + 4;
END;

PROCEDURE OpenScope;
VAR
  x: Object_;
BEGIN
  NEW(x);
  x.clas   := cHead;
  x.dsc    := topScope;
  x.c      := NIL;
  topScope := x;
END;

PROCEDURE CloseScope;
BEGIN
  topScope := topScope.dsc
END;

PROCEDURE NewObj(VAR x: Object_; id: Ident; clas: INTEGER);
VAR
  y: Object_;
BEGIN
  y := topScope;
  WHILE (y.c <> NIL) AND (y.c.name <> id) DO y := y.c;
  IF y.c = NIL THEN BEGIN
    NEW(x);
    x.name := id;
    x.clas := clas;
    x.op   := -1;
    x.c    := NIL;
    x.dsc  := NIL;
    y.c    := x;
  END ELSE BEGIN
    x := y.c;
    Mark('mult def');
  END;
END;

FUNCTION ThisObj(id: Ident): Object_;
VAR
  s, x: Object_;
BEGIN
  s := topScope;
  REPEAT
    x := s.c;
    WHILE (x <> NIL) AND (x.name <> id) DO x := x.c;
    s := s.dsc;
  UNTIL (x <> NIL) OR (s = NIL);
  RESULT := x;
END;

FUNCTION NewItem(clas: INTEGER; a, b: Item; op: INTEGER): Item;
VAR
  x: Item;
BEGIN
  NEW(x);
  x.clas := clas;
  x.typ  := noType;
  x.a    := a;
  x.b    := b;
  x.op   := op;
  x.val  := 0;
  x.reg  := 0;
  RESULT := x;
END;

PROCEDURE MakeConstItem(VAR x: Item; typ: Type_; val: INTEGER);
BEGIN
  x      := NewItem(cConst,NIL,NIL,0);
  x.typ  := typ;
  x.val  := val;
END;

PROCEDURE Enter(name: ARRAY OF CHAR; clas: INTEGER; typ: Type_; val: LONGINT);
VAR
  x: Object_;
BEGIN
  NEW(x);
  x.name  := COPY(name,0,length(name));
  x.clas  := clas;
  x.typ   := typ;
  x.val   := val;
  x.dsc   := NIL;
  x.a     := NIL;
  x.b     := NIL;
  x.c     := hidden;
  hidden  := x;
  IF clas = cTyp THEN typ.typobj := x;
END;

PROCEDURE Create(VAR x: Type_; form: INTEGER; size: LONGINT);
BEGIN
  NEW(x);
  x.size   := size;
  x.form   := form;
  x.base   := NIL;
  x.typobj := NIL;
END;

INITIALIZATION
  Mark     := @MarkProc;
  GC       := TObjectList.Create;

  hidden   := NIL;
  topScope := NIL;

  Create(byteType,fByte  , 1);
  Create(boolType,fBool  , 1);
  Create(charType,fChar  , 1);
  Create(intType ,fInt   , 4);
  Create(realType,fReal  , 4);
  Create(setType ,fSet   , 4);
  Create(nilType ,fNilTyp, 4);
  Create(noType  ,fNoTyp , 4);
  Create(hugeType,fHuge  , 8);
  Create(strType ,fString, 8);

  Enter('SET'     , cTyp, setType , 0);
  Enter('BYTE'    , cTyp, byteType, 0);
  Enter('CHAR'    , cTyp, charType, 0);
  Enter('REAL'    , cTyp, realType, 0);
  Enter('BOOLEAN' , cTyp, boolType, 0);
  Enter('INTEGER' , cTyp, intType , 0);
  Enter('HUGEINT' , cTyp, hugeType, 0);

  NEW(dummy);
  dummy.clas := cVars;
  dummy.typ  := intType;

  topScope   := NIL;
  OpenScope;
  topScope.c := hidden;

FINALIZATION
  GC.free;
END.

