UNIT MCE;

{$mode objfpc}{$H+}

INTERFACE

USES
  RTL, MCB, MCS, MCO;

VAR
  sym: INTEGER;

  PROCEDURE Check(s: INTEGER; msg: ARRAY OF CHAR);
  PROCEDURE Qualident(VAR x: MCB.Object_);
  PROCEDURE Expression(VAR x: MCB.Item);

IMPLEMENTATION

PROCEDURE Check(s: INTEGER; msg: ARRAY OF CHAR);
BEGIN
  IF sym = s THEN MCS.Get(sym) ELSE MCB.Mark(msg);
END;

PROCEDURE CheckTypes(op: INTEGER; VAR x, y: MCB.Item);
BEGIN
  x := MCB.NewItem(MCB.cNode,x,y,op);
END;

PROCEDURE Qualident(VAR x: MCB.Object_);
BEGIN
  x := MCB.ThisObj(MCS.id);
  MCS.Get(sym);
  IF x = NIL THEN BEGIN
    x := MCB.dummy;
    MCB.Mark('undef');
  END;
END;

PROCEDURE Factor(VAR x: MCB.Item);
BEGIN

  IF (sym < MCS.lx_char) AND (sym <> lx_minus) OR (sym > MCS.lx_ident) THEN BEGIN
    MCB.Mark('expression expected');
    REPEAT
      MCS.Get(sym)
    UNTIL (sym >= MCS.lx_char) AND (sym <= MCS.lx_ident) OR (sym = MCS.lx_eot);
  END;

  IF sym = MCS.lx_ident THEN BEGIN
    MCB.MakeConstItem(x,MCB.noType,0);
    x.name := MCS.id;
    x.clas := MCB.cLit;
    MCS.Get(sym);
  END ELSE IF sym = MCS.lx_int    THEN BEGIN MCB.MakeConstItem(x,MCB.intType,MCS.i32);  MCS.Get(sym);
  END ELSE IF sym = MCS.lx_real   THEN BEGIN MCB.MakeConstItem(x,MCB.realType,MCS.i32); MCS.Get(sym);
  END ELSE IF sym = MCS.lx_lparen THEN BEGIN MCS.Get(sym); Expression(x); Check(MCS.lx_rparen,'rparen ?');
  END ELSE IF sym = MCS.lx_minus  THEN BEGIN MCS.Get(sym); Factor(x); x.op := MCS.lx_not;
  END ELSE BEGIN
    MCB.Mark('not a factor');
    MCB.MakeConstItem(x,MCB.intType,0);
  END;

END;

PROCEDURE Term(VAR x: MCB.Item);
VAR
  y : MCB.Item;
  op: INTEGER;
BEGIN
  Factor(x);
  WHILE (sym >= MCS.lx_times) AND (sym <= MCS.lx_and) DO BEGIN
    op := sym;
    MCS.Get(sym);
    Factor(y);
    CheckTypes(op,x,y);
  END;
END;

PROCEDURE SimpleExpression(VAR x: MCB.Item);
VAR
  y : MCB.Item;
  op: INTEGER;
BEGIN
  IF sym = MCS.lx_plus THEN BEGIN
    MCS.Get(sym);
    Term(x);
  END ELSE
    Term(x);
  WHILE (sym >= MCS.lx_plus) AND (sym <= MCS.lx_or) DO BEGIN
    op := sym;
    MCS.Get(sym);
    Term(y);
    CheckTypes(op,x,y);
  END;
END;

PROCEDURE Expression(VAR x: MCB.Item);
VAR
  y, z: MCB.Item;
  op  : INTEGER;
BEGIN
  SimpleExpression(x);
  IF (sym >= MCS.lx_eql) AND (sym <= MCS.lx_geq) THEN BEGIN
    op := sym;
    MCS.Get(sym);
    SimpleExpression(y);
    CheckTypes(op,x,y);
    z.typ := MCB.boolType;
  END;
  MCO.Simplification(x);
  //MCO.ItemMark(x);
END;

END.

