UNIT MCO;

{$mode objfpc}{$H+}

INTERFACE

USES
  RTL, MCB, MCS;

  PROCEDURE Simplification(VAR x: MCB.Item);
  PROCEDURE ItemMark(x: MCB.Item);

IMPLEMENTATION

PROCEDURE Simplification(VAR x: MCB.Item);
VAR
  found: BOOLEAN;

  PROCEDURE Swap(VAR x,y: MCB.Item);
  VAR
    z: MCB.Item;
  BEGIN
    z := x;
    x := y;
    y := z;
  END;

  PROCEDURE SwapOp(VAR x,y: MCB.Item);
  VAR
    op: INTEGER;
  BEGIN
    op   := x.op;
    x.op := y.op;
    y.op := op;
  END;

  PROCEDURE Multiplier(VAR x: MCB.Item);
  VAR
    y,z: MCB.Item;
  BEGIN
    y := x;
    z := NIL;
    WHILE (y.a<>NIL) AND (y.a.clas = MCB.cNode) DO BEGIN
      IF (y.op IN [MCS.lx_times]) AND (y.b.clas = MCB.cConst) THEN z := y;
      IF (z <> NIL) THEN BEGIN
        IF NOT (y.a.op IN [MCS.lx_plus,MCS.lx_minus]) AND (y.a.val <> - 1) THEN BEGIN
          Swap(z.b,y.a.b);
          SwapOp(y.a,z);
          z := y.a;
        END;
      END;
      y      := y.a;
    END;
    IF (z <> NIL) THEN BEGIN
      IF z.a.op IN [MCS.lx_rdiv] THEN BEGIN
        Swap(z.a,z.b);
      END ELSE IF (z.a.clas = MCB.cConst) AND (z.a.a = NIL) AND (z.b.clas = MCB.cConst) THEN BEGIN
        z.val := -1;
      END ELSE IF (z.a.clas <> MCB.cConst) AND (z.b.clas = MCB.cConst) THEN BEGIN
        Swap(z.a,z.b);
        z.val := 0;
      END;
    END ELSE
      x.val := - 1;
  END;

  PROCEDURE Summand(VAR x: MCB.Item);
  VAR
    y,z: MCB.Item;
  BEGIN

    y := x;
    z := NIL;
    WHILE (y.a<>NIL) AND (y.a.clas = MCB.cNode) DO BEGIN
      IF (y.op IN [MCS.lx_plus,MCS.lx_minus]) AND (y.b.clas = MCB.cConst) THEN z := y;
      IF (z <> NIL) THEN BEGIN
        IF NOT (y.a.op IN [MCS.lx_times,MCS.lx_rdiv,MCS.lx_mod,MCS.lx_div]) AND (y.a.val <> - 2) THEN BEGIN
          Swap(z.b,y.a.b);
          SwapOp(y.a,z);
          z := y.a;
        END;
      END;
      y := y.a;
    END;
    IF (z <> NIL) THEN BEGIN
      IF z.a.op IN [MCS.lx_times,MCS.lx_rdiv,MCS.lx_mod,MCS.lx_div] THEN BEGIN
        Swap(z.a,z.b);
      END ELSE IF (z.a.clas = MCB.cConst) AND (z.a.a = NIL) AND (z.b.clas = MCB.cConst) THEN BEGIN
        z.val := -2;
      END ELSE IF (z.a.clas <> MCB.cConst) AND (z.b.clas = MCB.cConst) THEN BEGIN
        IF z.op = MCS.lx_minus THEN BEGIN z.b.op := MCS.lx_not; z.op := MCS.lx_plus; END;
        Swap(z.a,z.b);
        z.val := 0;
      END;
    END ELSE
      x.val := -2;

  END;

BEGIN
  WHILE (x.a <> NIL) AND (x.val <> -1) DO Multiplier(x);
  WHILE (x.a <> NIL) AND (x.val <> -2) DO Summand(x);
  IF x.a <> NIL THEN Simplification(x.a);
  IF x.b <> NIL THEN Simplification(x.b);

  IF (*FALSE AND*) (x.a <> NIL) AND (x.b <> NIL) THEN BEGIN
    IF (x.a.clas = MCB.cConst) AND (x.b.clas = MCB.cConst) THEN BEGIN

      IF x.a.op = MCS.lx_not THEN x.a.val := - x.a.val;
      IF x.b.op = MCS.lx_not THEN x.b.val := - x.b.val;

      found := TRUE;
      CASE x.op OF
        MCS.lx_plus  : x.val := x.a.val + x.b.val;
        MCS.lx_minus : x.val := x.a.val - x.b.val;
        MCS.lx_rdiv  : x.val := x.a.val div x.b.val;
        MCS.lx_times : x.val := x.a.val * x.b.val;
        MCS.lx_mod   : x.val := RTL._MOD(x.a.val,x.b.val);
        MCS.lx_div   : x.val := RTL._DIV(x.a.val,x.b.val);
      ELSE
        found := FALSE;
      END;

      IF found THEN BEGIN
        x.typ  := x.b.typ;
        x.clas := MCB.cConst;
        x.op   := MCS.lx_null;
        x.a    := NIL;
        x.b    := NIL;
        x.reg  := -1;
      END;

    END;
  END;

END;

PROCEDURE ItemMark(x: MCB.Item);

  FUNCTION Max(r1,r2: INTEGER): INTEGER;
  BEGIN
    RESULT := r1;
    IF r1 < r2 THEN RESULT := r2;
  END;

BEGIN
  IF (x <> NIL) AND (x.a <> NIL) AND (x.b <> NIL) THEN BEGIN
    IF x.clas = MCB.cNode THEN BEGIN
      x.reg := 1;
    END ELSE IF x.a.clas = MCB.cNode THEN BEGIN
      ItemMark(x.a);
      x.reg := x.a.reg;
    END ELSE BEGIN
      ItemMark(x.a);
      ItemMark(x.b);
      IF x.a.reg = x.b.reg THEN
        x.reg := x.a.reg + 1
      ELSE
        x.reg := Max(x.a.reg,x.b.reg);
    END;
  END;
END;

END.

