UNIT RTL;

{$mode objfpc}{$H+}

INTERFACE

  FUNCTION _DIV(CONST x, y: INT64): INT64;
  FUNCTION _MOD(CONST x, y: INT64): INT64;
  FUNCTION _ASH (e, bit : INTEGER)  : INTEGER;
  FUNCTION _LSL(CONST x, y: INTEGER): INTEGER;
  FUNCTION _ASR(CONST x, y: INTEGER): INTEGER;
  FUNCTION _ROR(CONST x, y: INTEGER): INTEGER;
  FUNCTION _FLT(CONST i: INTEGER)   : SINGLE;
  FUNCTION _ITR(CONST i: INTEGER)   : SINGLE;
  FUNCTION _RTI(CONST r: SINGLE)    : INTEGER;
  FUNCTION _HTR(CONST i: INT64)     : DOUBLE;
  FUNCTION _RTH(CONST r: DOUBLE)    : INT64;
  FUNCTION _AND(CONST x, y: INTEGER): INTEGER;
  FUNCTION _HAH(p: POINTER; l: INTEGER): INTEGER;

IMPLEMENTATION

FUNCTION _DIV(CONST x, y: INT64): INT64;
VAR
  Q,R: INT64;
BEGIN
  Q:=TRUNC(x/y);
  R:=x MOD y;
  IF R<0 THEN BEGIN
    IF y>0 THEN
      Q:=Q-1
    ELSE
      Q:=Q+1;
  END;
  RESULT := Q;
END;

FUNCTION _MOD(CONST x, y: INT64): INT64;
BEGIN
  ASSERT(y > 0);
  IF x >= 0 THEN
    RESULT := x MOD y
  ELSE
    RESULT := (y + (x MOD y)) MOD y;
END;

FUNCTION _ASH (e, bit : INTEGER): INTEGER;
BEGIN
  IF bit < 0 THEN e := e SHR SYSTEM.ABS(bit) ELSE e := e SHL bit;
  RESULT := e;
END;

FUNCTION _LSL(CONST x, y: INTEGER): INTEGER;
BEGIN
  RESULT := x SHL (y AND 31);
END;

FUNCTION _ASR(CONST x, y: INTEGER): INTEGER;
BEGIN
  RESULT := (LONGWORD(x)) SHR (y AND 31);
END;

FUNCTION _ROR(CONST x, y: INTEGER): INTEGER;
BEGIN
  RESULT := (x SHR (y AND 31)) OR (x SHL (-y AND 31));
END;

FUNCTION _FLT(CONST i: INTEGER): SINGLE;
BEGIN
  RESULT := i*1.0;
END;

FUNCTION _ITR(CONST i: INTEGER): SINGLE;
VAR
  r: SINGLE ABSOLUTE i;
BEGIN
  RESULT := r;
END;

FUNCTION _RTI(CONST r: SINGLE): INTEGER;
VAR
  i: INTEGER ABSOLUTE r;
BEGIN
  RESULT := i;
END;

FUNCTION _HTR(CONST i: INT64): DOUBLE;
VAR
  r: DOUBLE ABSOLUTE i;
BEGIN
  RESULT := r;
END;

FUNCTION _RTH(CONST r: DOUBLE): INT64;
VAR
  i: INT64 ABSOLUTE r;
BEGIN
  RESULT := i;
END;

FUNCTION _AND(CONST x, y: INTEGER): INTEGER;
BEGIN
  RESULT := x AND y;
END;

FUNCTION _HAH(p: POINTER; l: INTEGER): INTEGER;
VAR
  i: POINTER;
  x: INTEGER;
BEGIN
  I := P;
  RESULT := 0;
  x := 0;
  WHILE x < L DIV 2 DO BEGIN
    RESULT := RESULT * 1664525 + BYTE(I^) + 1013904223;
    INC(I,2);
    INC(x);
  END;
END;

END.

